# variable declarations

model = {
    "id": "Flight Track",
    "name": "ER2",
    "availability": "{}/{}",
    "model": {
        "gltf": "https://s3.amazonaws.com/visage-czml/iphex_HIWRAP/img/er2.gltf",
        "scale": 100.0,
        "minimumPixelSize": 32,
        "maximumScale": 150.0
    },
    "position": {
        "cartographicDegrees": []
    },
    "path": {
        "material": {
            "solidColor": {
                "color": {
                    "rgba": [0, 255, 128, 255]
                }
            }
        },
        "width": 1,
        "resolution": 5
    },
    "properties": {
        "roll": {},
        "pitch": {},
        "heading": {}
    }
}

czml_head = {
    "id": "document",
    "name": "wall czml",
    "version": "1.0"
}


# abc: abstract base class
from abc import ABC, abstractmethod

class DataProcess(ABC):
  
  @abstractmethod
  def ingest(self):
    # print('get data from various sources to be preprocessed and visualized in Cesium')
    pass
  
  @abstractmethod
  def preprocess(self):
    # print('with the data, preprocess it to be ready for vizualization')
    # print('can go through multiple steps of preprocessing like cleaning, transformation, integration.')
    pass
  
  @abstractmethod
  def prep_visualization(self):
    # print('After preprocessing, prepare data to be visualized in Cesium')
    pass


class CesiumViz():
  def formulate_body(self, script: str) -> str:
    # print('Formaulate a predefined HTML BODY structure with the custom script')
    return \
      f"""
          <!DOCTYPE html>
          <html lang="en">
          <head>
            <meta charset="utf-8">
            <!-- Include the CesiumJS JavaScript and CSS files -->
            <script src="https://cesium.com/downloads/cesiumjs/releases/1.96/Build/Cesium/Cesium.js"></script>
            <script src="https://releases.jquery.com/git/jquery-git.min.js"></script>
            <link href="https://cesium.com/downloads/cesiumjs/releases/1.96/Build/Cesium/Widgets/widgets.css" rel="stylesheet">
            </head>
          <body>
            <div id="cesiumContainer"></div>
              <script type="module">
                {script}
              </script>
            </div>
          </body>
          </html>
    """

  def add_script(self, url: str) -> str:
    # print('add script that creates cesium viewer and loads it with various kinds of data (czml, 3dtilesets, etc))')
    pass

  def generate_html(self, url: str) -> str:
    # print('Generate html file with cesium script and body')
    script = self.add_script(url)
    html = self.formulate_body(script)
    # print("HTML generation complete.")
    return html


from abc import abstractmethod
import numpy as np
import pandas as pd

from .data_process import DataProcess

class TilesPointCloudDataProcess(DataProcess):
  @abstractmethod
  def ingest(self, url: str) -> pd.DataFrame:
    # print("get raw data from data source. Data source may be any of http or ftp")
    # raw data can be in any file format. eg. txt, nc, uf etc.
    # use appropriate library to read the file and return the data as pandas dataframe
    pass

  @abstractmethod
  def preprocess(self):
    # print("preprocess raw data to be ready for vizualization")
    # pre process of raw data can go through various stages like
    # cleaning: scrape the necessary data columns,
    # transformation: transform the data to a suitable data formatting
    # integration: data from multiple sources can be integrated into intermediate file format, e.g. zarr file. The intermeidate format should be compatible with viz prepration step
    pass
    
  @abstractmethod
  def prep_visualization(self):
    # print("After preprocessing, prepare tileset.json data to be visualized in Cesium using tileset writer")
    # tileset writer a util class can be used to write tileset json data from zarr file as input
    pass


import boto3
from copy import deepcopy
import numpy as np
import pandas as pd
from typing import Generator

from .abstract.czml_data_process import CZMLDataProcess
from .utils.czml_writer_nav import NavCzmlWriter

class NavCZMLDataProcess(CZMLDataProcess):
  def __init__(self):    
    pass
  
  def ingest(self, url: str, type: str = "local") -> np.array:
    """Returns a numpy array representing the raw data file.

    Keyword arguments:
    url -- the path of the raw data file.
    type -- either s3 or local (default local).
    """
    if (type == "s3"):
      return self._ingest_from_s3(url)
    else:
      return self._ingest_from_local(url)
  
  def preprocess(self, data: np.array) -> pd.DataFrame:
    cleaned_data = self._cleaning(data)
    transformed_data = self._transformation(cleaned_data)
    integrated_data = self._integration(transformed_data)
    return integrated_data

  def prep_visualization(self, data: pd.DataFrame) -> str:
    # the dataframe needs these features/cols: timestamp, longitude, latitude, altitude, roll, pitch, heading
    length = data.shape[0]
    nav_czml_writer = NavCzmlWriter(length)
    nav_czml_writer.set_with_df(data)
    nav_czml_str = nav_czml_writer.get_czml_string()
    return nav_czml_str


  # data preprocessing steps

  def _cleaning(self, data: np.array) -> pd.DataFrame:
    col_index_map = self._get_col_index_map()
    data = deepcopy(data)

    # data extraction
    # scrape necessary data columns 
    time = data[:, col_index_map["time"]]
    latitude = data[:, col_index_map["latitude"]]
    longitude = data[:, col_index_map["longitude"]]
    altitude = data[:, col_index_map["altitude"]]
    heading = data[:, col_index_map["heading"]] * np.pi / 180. - np.pi / 2.
    pitch = data[:, col_index_map["pitch"]] * np.pi / 180.
    roll = data[:, col_index_map["roll"]] * np.pi / 180.
    
    # data masks
    # remove nan values
    mask = np.logical_not(np.isnan(latitude))
    mask = np.logical_and(mask, np.logical_not(np.isnan(longitude)))
    mask = np.logical_and(mask, np.logical_not(np.isnan(altitude)))
    mask = np.logical_and(mask, np.logical_not(np.isnan(heading)))
    mask = np.logical_and(mask, np.logical_not(np.isnan(pitch)))
    mask = np.logical_and(mask, np.logical_not(np.isnan(roll)))
    
    # remove duplicate time values
    _, unique_idx = np.unique(time, return_index=True)
    unique = np.copy(mask)
    unique[:] = False
    unique[unique_idx] = True
    mask = np.logical_and(mask, unique)
    
    # apply masks/filter and sample data
    f_time = time[mask].astype('datetime64[s]')[::5]
    f_latitude = latitude[mask][::5]
    f_longitude = longitude[mask][::5]
    f_altitude = altitude[mask][::5]
    f_heading = heading[mask][::5]
    f_pitch = pitch[mask][::5]
    f_roll = roll[mask][::5]

    filtered_data = pd.DataFrame(data = {"timestamp": f_time, "latitude": f_latitude, "longitude": f_longitude, "altitude": f_altitude, "heading": f_heading, "pitch": f_pitch, "roll": f_roll})
    return filtered_data
  
  def _transformation(self, data: pd.DataFrame) -> pd.DataFrame:
    #  no transformation needed
    return data
  
  def _integration(self, data: pd.DataFrame) -> pd.DataFrame:
    # no integration needed
    return data


  # ingestion variations

  def _ingest_from_s3(self, url: str) -> np.array:
    s3_client = boto3.client('s3')
    [bucket_name, objectKey] = self._get_s3_details(url)
    s3_file = s3_client.get_object(Bucket=bucket_name, Key=objectKey)

    file = s3_file['Body'].iter_lines()

    data = self._generator_to_np(file)
    return data

  def _ingest_from_local(self, path: str) -> np.array:
    data = self._generator_to_np(path)
    return data


  # utils

  def _get_s3_details(self, url) -> list:
    url = url.replace("s3://", "")
    temp_url = url.split("/")
    bucket_name = temp_url[0]
    
    temp_url = url.split(bucket_name+"/")
    objectKey = temp_url[1]  # key should not start with /
    return [bucket_name, objectKey]
  
  def _get_col_index_map(self):
    # represents the column number for each key, inside the input csv type file.
    return {
        "time": 1,
        "latitude": 2,
        "longitude": 3,
        "altitude": 4,
        "heading": 14,
        "pitch": 16,
        "roll": 17
    }
    
  def _generator_to_np(self, infile: Generator) -> np.array:
    # As the data in txt is all string, to put it inside numpy array, we need to convert it to appropirate types
    
    # create null converters
    converters = {}
    for i in range(33): # 33 cols/feature data
      converters[i] = lambda _: np.nan 
    
    # upadate converters for appropriate faeture/cols with appropriate functions
    col_index_map = self._get_col_index_map()
    converters[col_index_map["time"]] = lambda x: np.datetime64(x, 's').astype(np.int64)
    converters[col_index_map["latitude"]] = self._string_to_float
    converters[col_index_map["longitude"]] = self._string_to_float
    converters[col_index_map["altitude"]] = self._string_to_float
    converters[col_index_map["heading"]] = self._string_to_float
    converters[col_index_map["pitch"]] = self._string_to_float
    converters[col_index_map["roll"]] = self._string_to_float
    
    # apply converter during txt load
    return np.loadtxt(infile, delimiter=',', converters=converters)
  
  def _string_to_float(self, str: str) -> np.float64:
        value = np.nan
        try:
            value = float(str)
        except:
            pass
        return value


from .base_viz import CesiumViz

class CZMLViz(CesiumViz):
  def add_script(self, czml_path: str) -> str:
    # print('add script that creates cesium viewer and loads it with various kinds of data (czml, 3dtilesets, etc))')
    s1= \
      """
      Cesium.Ion.defaultAccessToken = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJqdGkiOiJhYTg3MTQxNS01ZTIwLTQ4MmItYjA5NS1hYWM3MWQ0OTNkYTMiLCJpZCI6MTA1ODMzLCJpYXQiOjE2NjE0MzkwOTB9.IWdoSi1zjC7fl7Ncj0YVJgXMfjX3K-RmRcGtjp2xryo";
                
      const viewer = new Cesium.Viewer("cesiumContainer", {
        terrainProvider: Cesium.createWorldTerrain(),
        shouldAnimate: true,
      });
    
      loadCZMLData()
    
      function loadCZMLData() {
      """
    s2 = \
      f"""
        Cesium.CzmlDataSource.load("{czml_path}")
      """
    s3 = \
      """
        .then(async (dataSource) => {
          viewer.dataSources.add(dataSource);
          // zoom to entity, SET CAMERA ANGLE
          viewer.zoomTo(dataSource,  new Cesium.HeadingPitchRange(0, Cesium.Math.toRadians(-10), 40000));

          // only apply to czml for flight entities
          const flightEntity = dataSource.entities.getById("Flight Track");
          if (flightEntity) {
              // set camera view
              flightEntity.viewFrom = new Cesium.Cartesian3(-30000, -70000, 50000);
              
              // set camera to track entity
              viewer.trackedEntity = flightEntity;

              // fix orientation
              flightEntity.orientation = new Cesium.CallbackProperty((time, _result) => fixOrientation(flightEntity, time), false);

              function fixOrientation(entity, time) {
                  const position = entity.position.getValue(time);
                  let { heading, pitch, roll, correctionOffsets } = entity.properties.getValue(time);
                  // only the heading should change with respect to the position.
                  if(!correctionOffsets) {
                      correctionOffsets = {
                          heading: 0,
                          pitch: 0,
                          roll: 0
                      }
                  }
                  // fix the pitch and roll rotations
                  heading = heading + Cesium.Math.toRadians(correctionOffsets.heading);
                  pitch = pitch + Cesium.Math.toRadians(correctionOffsets.pitch);
                  roll = roll + Cesium.Math.toRadians(correctionOffsets.roll);
                  const hpr = new Cesium.HeadingPitchRoll(heading, pitch, roll);
                  const fixedOrientation = Cesium.Transforms.headingPitchRollQuaternion(
                      position,
                      hpr
                  );
                  return fixedOrientation;
              }
          }
        });
      }
      """
    return s1 + s2 + s3


from abc import abstractmethod
import numpy as np
import pandas as pd

from .data_process import DataProcess

class CZMLDataProcess(DataProcess):
  @abstractmethod
  def ingest(self, url: str) -> np.array:
    # print("get raw data from data source. Data source may be any of http or ftp")
    # raw data can be in any file format. eg. txt, nc, uf etc.
    # use appropriate library to read the file and return the data as numpy arrays
    pass

  @abstractmethod
  def preprocess(self) -> pd.DataFrame:
    # print("preprocess raw data to be ready for vizualization")
    # pre process of raw data can go through various stages like
    # cleaning: scrape the necessary data columns,
    # transformation: transform the data to a suitable data formatting
    # integration: data from multiple sources can be integrated into intermediate format, e.g. numpy arrays. The intermeidate format should be compatible with viz prepration step
    pass
    
  @abstractmethod
  def prep_visualization(self) -> str:
    # print("After preprocessing, prepare czml data to be visualized in Cesium using CZML writer")
    # czml writer a util class can be used to write czml data from numpy arrays as input
    pass


import numpy as np
import json

from copy import deepcopy
from .czml_model import model, czml_head

class NavCzmlWriter:
    """
    A class for generating CZML (Cesium Markup Language) data for navigation visualization.

    Args:
        length (int): The length of data points to be generated.
        time_window (tuple): A tuple containing the start and end time of the data.
        time_steps (numpy.ndarray): An array of time steps for the generated data.
        longitude (numpy.ndarray): An array of longitudes for the generated data.
        latitude (numpy.ndarray): An array of latitudes for the generated data.
        altitude (numpy.ndarray): An array of altitudes for the generated data.
        roll (numpy.ndarray): An array of roll values for the generated data.
        pitch (numpy.ndarray): An array of pitch values for the generated data.
        heading (numpy.ndarray): An array of heading values for the generated data.

    Attributes:
        model (dict): A CZML model containing position and orientation data.

    Methods:
        __init__(length, time_window, time_steps, longitude, latitude, altitude, roll, pitch, heading):
            Initializes the NavCzmlWriter instance and sets up the CZML model.
        
        _set_with_df(df):
            Sets the CZML data using a DataFrame containing navigation data.

        _set_time(time_window, time_steps):
            Sets the time-related properties in the CZML model.
        
        _set_position(longitude, latitude, altitude):
            Sets the position-related properties in the CZML model.
        
        _set_orientation(roll, pitch, heading):
            Sets the orientation-related properties in the CZML model.
        
        _get_time_info(time):
            Converts the time data into CZML-compatible time window and time steps.
        
        get_czml_string():
            Returns the CZML data as a JSON string.
    """
    def __init__(self, length):
    # def __init__(self, length):
        self.model = deepcopy(model)
        self.czml_head = deepcopy(czml_head)
        self.length = length
        self.model['position']['cartographicDegrees'] = [0] * 4 * length
        self.model['properties']['roll']['number'] = [0] * 2 * length
        self.model['properties']['pitch']['number'] = [0] * 2 * length
        self.model['properties']['heading']['number'] = [0] * 2 * length

    def set_with_df(self, df):
        self._set_time(*self._get_time_info(df['timestamp']))
        self._set_position(df['longitude'], df['latitude'], df['altitude'])
        self._set_orientation(df['roll'], df['pitch'], df['heading'])

    def _set_time(self, time_window, time_steps):
        [epoch, end] = time_window
        self.model['availability'] = f"{epoch}/{end}"
        self.model['position']['epoch'] = epoch
        self.model['position']['cartographicDegrees'][0::4] = time_steps
        self.model['properties']['roll']['epoch'] = epoch
        self.model['properties']['pitch']['epoch'] = epoch
        self.model['properties']['heading']['epoch'] = epoch
        self.model['properties']['roll']['number'][0::2] = time_steps
        self.model['properties']['pitch']['number'][0::2] = time_steps
        self.model['properties']['heading']['number'][0::2] = time_steps

    def _set_position(self, longitude, latitude, altitude):
        self.model['position']['cartographicDegrees'][1::4] = longitude
        self.model['position']['cartographicDegrees'][2::4] = latitude
        self.model['position']['cartographicDegrees'][3::4] = altitude

    def _set_orientation(self, roll, pitch, heading):
        self.model['properties']['roll']['number'][1::2] = roll
        self.model['properties']['pitch']['number'][1::2] = pitch
        self.model['properties']['heading']['number'][1::2] = heading

    def _get_time_info(self, time):
        time = time.values.astype('datetime64[s]') # pandas series to numpy ndarray
        time_window = time[[0, -1]].astype(np.string_)  # get first and last element
        time_window = np.core.defchararray.add(time_window, np.string_('Z')) # add Z to each time window element to make it ISO format
        time_window = np.core.defchararray.decode(time_window, 'UTF-8') # decode to UTF-8 from byte_ object
        time_steps = (time - time[0]).astype(int).tolist()
        return time_window, time_steps

    def get_czml_string(self):
        return json.dumps([self.czml_head, self.model])



tileset_json = {
	"asset": {
		"version": "1.0",
		"type": "Airborne Radar"
	},
	"root": {
		"geometricError": 1000000,
		"refine" : "REPLACE",
		"boundingVolume": {
            "region": []
        },
        "children": []
	},
    "properties": {
        "epoch": "",
        "refined": []
    }
}


from .base_viz import CesiumViz

class TilesViz(CesiumViz):
  def add_script(self, tileset_path: str) -> str:
    p1= \
      """
      Cesium.Ion.defaultAccessToken = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJqdGkiOiJhYTg3MTQxNS01ZTIwLTQ4MmItYjA5NS1hYWM3MWQ0OTNkYTMiLCJpZCI6MTA1ODMzLCJpYXQiOjE2NjE0MzkwOTB9.IWdoSi1zjC7fl7Ncj0YVJgXMfjX3K-RmRcGtjp2xryo";
                
      const viewer = new Cesium.Viewer("cesiumContainer", {
        terrainProvider: Cesium.createWorldTerrain(),
        shouldAnimate: true,
      });

      load3DTileData()
    
      function load3DTileData() {
        const tileset = new Cesium.Cesium3DTileset({

      """
    p2 = \
      f"""
        url: "{tileset_path}"
      """
    p3 = \
      """
        });

        tileset.style = new Cesium.Cesium3DTileStyle({
            color: getColorExpression(),
            pointSize: 5.0
          });

        var currentTime = Cesium.JulianDate.fromIso8601("2015-11-10T17:54:00Z")
        var endTime = Cesium.JulianDate.fromIso8601("2015-11-10T23:59:00Z");

        viewer.clock.currentTime = currentTime;
        viewer.clock.multiplier = 10;

        // add the tileset to viewer.
        viewer.scene.primitives.add(tileset);

        viewer.timeline.zoomTo(currentTime, endTime);
        viewer.zoomTo(tileset);
      }
              
      function getColorExpression() {
        // helper to generate color expression,
        //  to style the tileset refer. tilset styling in cesium docs for more info on styling expressions. 
          let reverse = true
          let ascale = 4.346
          let vmin = -10
          let vmax = 30
          let vrange = vmax - vmin
          let hmin = 0.438
          let hrange = 1

          let revScale = ""
          if (reverse) {
              revScale = " * -1.0 + 1.0"
          }
          return `hsla((((clamp(\${value}, ${vmin}, ${vmax}) + ${vmin}) / ${vrange}) ${revScale}) * ${hrange} + ${hmin}, 1.0, 0.5, pow((\${value} - ${vmin})/${vrange}, ${ascale}))`
      }
      """
    return p1 + p2 + p3


import os
import shutil
import boto3
import numpy as np
import pandas as pd
import xarray as xr
import zarr

from typing import Generator

from .abstract.tiles_pointcloud_data_process import TilesPointCloudDataProcess
from .utils.tiles_writer import write_tiles

class RadRangeTilesPointCloudDataProcess(TilesPointCloudDataProcess):
  def __init__(self):
    self.url = None
    self.OPENED_FILE_REF = None

    self.campaign = 'Olympex'
    self.collection = "AirborneRadar"
    self.dataset = "gpmValidationOlympexcrs"
    self.variables = ["zku"]
    self.renderers = ["point_cloud"]

    self.chunk = 262144
    self.to_rad = np.pi / 180
    self.to_deg = 180 / np.pi

  def ingest(self, url: str, type: str = "local") -> xr.Dataset:
    """Returns a numpy array representing the raw data file.

    Keyword arguments:
    url -- the path of the raw data file.
    type -- either s3 or local (default local).
    """
    if (type == "s3"):
      return self._ingest_from_s3(url)
    else:
      return self._ingest_from_local(url)
  
  def preprocess(self, data: xr.Dataset) -> str:
    cleaned_data = self._cleaning(data)
    transformed_data = self._transformation(cleaned_data)
    integrated_data = self._integration(transformed_data)
    self.OPENED_FILE_REF.close()
    return integrated_data

  def prep_visualization(self, zarr_data_path: str) -> str:
    point_cloud_folder = zarr_data_path+"_point_cloud"
    write_tiles("ref", 0, 1000000000000, zarr_data_path, point_cloud_folder)
    return point_cloud_folder

  # data preprocessing steps

  def _cleaning(self, data: xr.Dataset) -> xr.Dataset:
    # data extraction
    # scrape necessary data columns 
    extracted_data = data[['timed', 'zku', 'lat', 'lon', 'altitude', 'roll', 'pitch', 'head', 'range']]
    return extracted_data

  def _transformation(self, data: xr.Dataset) -> pd.DataFrame:
    #  transform the data to a suitable data formatting
    hour = data['timed'].values
    lat = data['lat'].values
    lon = data['lon'].values
    alt = data['altitude'].values # altitude of aircraft in meters
    roll = data["roll"].values
    pitch = data["pitch"].values
    head = data["head"].values
    ref = data['zku'].values #CRS radar reflectivity #2d data
    rad_range = data["range"].values # has lower count than ref
    
    # time correction and conversion:
    base_time = self._get_date_from_url(self.url)
    hour = self._add24hr(hour)
    delta = (hour * 3600).astype('timedelta64[s]') + base_time
    time = (delta - np.datetime64('1970-01-01')).astype('timedelta64[s]').astype(np.int64)

    # transform ref to 1d array and repeat other columns to match data dimension

    num_col = ref.shape[0] # number of cols
    num_row = ref.shape[1] # number of rows

    time = np.repeat(time, num_row)
    lon = np.repeat(lon, num_row)
    lat = np.repeat(lat, num_row)
    alt = np.repeat(alt, num_row)
    roll = np.repeat(roll * self.to_rad, num_row)
    pitch = np.repeat(pitch * self.to_rad, num_row)
    head = np.repeat(head * self.to_rad, num_row)
    rad_range = np.tile(rad_range, num_col)
    ref = ref.flatten()

    # curtain creation

    x, y, z = self._down_vector(roll, pitch, head)
    x = np.multiply(x, np.divide(rad_range, 111000 * np.cos(lat * self.to_rad)))
    y = np.multiply(y, np.divide(rad_range, 111000))
    z = np.multiply(z, rad_range)
    lon = np.add(-x, lon)
    lat = np.add(-y, lat)
    alt = np.add(z, alt)

    # sort by time

    sort_idx = np.argsort(time)
    lon = lon[sort_idx]
    lat = lat[sort_idx]
    alt = alt[sort_idx]
    ref = ref[sort_idx]
    time = time[sort_idx]

    # remove nan and infinite using mask (dont use masks filtering for values used for curtain creation)

    mask = np.logical_and(np.isfinite(ref), alt > 0)
    time = time[mask]
    ref = ref[mask]
    lon = lon[mask]
    lat = lat[mask]
    alt = alt[mask]


    df = pd.DataFrame(data = {
      'time': time,
      'lon': lon,
      'lat': lat,
      'alt': alt,
      'ref': ref
    })

    return df
  
  def _integration(self, data: pd.DataFrame) -> str:
    # data from multiple sources can be integrated into intermediate file format, e.g. zarr file. The intermeidate format should be compatible with viz prepration step
    time = data['time'].values
    ref = data['ref'].values
    lon = data['lon'].values
    lat = data['lat'].values
    alt = data['alt'].values

    # path creation
    zarr_path = self._create_zarr_dir()

    # create a ZARR directory in the path provided
    store = zarr.DirectoryStore(zarr_path)
    root = zarr.group(store=store)

    # Create empty rows for modified data inside zarr
    z_chunk_id = root.create_dataset('chunk_id', shape=(0, 2), chunks=None, dtype=np.int64)
    z_location = root.create_dataset('location', shape=(0, 3), chunks=(self.chunk, None), dtype=np.float32)
    z_time = root.create_dataset('time', shape=(0), chunks=(self.chunk), dtype=np.int32)
    z_vars = root.create_group('value')
    z_ref = z_vars.create_dataset('ref', shape=(0), chunks=(self.chunk), dtype=np.float32)
    n_time = np.array([], dtype=np.int64)

    # Now populate (append) the empty rows in ZARR dir with preprocessed data
    z_location.append(np.stack([lon, lat, alt], axis=-1))
    z_ref.append(ref)
    n_time = np.append(n_time, time)

    idx = np.arange(0, n_time.size, self.chunk)
    chunks = np.zeros(shape=(idx.size, 2), dtype=np.int64)
    chunks[:, 0] = idx
    chunks[:, 1] = n_time[idx]
    z_chunk_id.append(chunks)

    epoch = np.min(n_time)
    n_time = (n_time - epoch).astype(np.int32)
    z_time.append(n_time)

    # save it.
    root.attrs.put({
        "campaign": self.campaign,
        "collection": self.collection,
        "dataset": self.dataset,
        "variables": self.variables,
        "renderers": self.renderers,
        "epoch": int(epoch)
    })

    return zarr_path

  # ingesting variations

  def _ingest_from_local(self, path: str) -> xr.Dataset:
    self.url = path
    data = self._generator_to_xr(path)
    return data

  def _ingest_from_s3(self, url: str) -> xr.Dataset:
    self.url = url
    s3_client = boto3.client('s3')
    [bucket_name, objectKey] = self._get_s3_details(url)

    s3_file = s3_client.get_object(Bucket=bucket_name, Key=objectKey)
    file = s3_file['Body'].read()

    data = self._generator_to_xr(file)
    return data

  # utils

  def _get_s3_details(self, url) -> list:
    url = url.replace("s3://", "")
    temp_url = url.split("/")
    bucket_name = temp_url[0]
    
    temp_url = url.split(bucket_name+"/")
    objectKey = temp_url[1]  # key should not start with /
    return [bucket_name, objectKey]
  
  def _generator_to_xr(self, infile: Generator) -> xr.Dataset:
    # As the data in netcdf format, to put it inside numpy array, we need to read it first.
    ds = xr.open_dataset(infile, decode_cf=False) # dont close opened dataset just yet
    self.OPENED_FILE_REF = ds # close later, after data preprocessing.
    return ds
  
  def _add24hr(self, hour: np.ndarray) -> np.ndarray:
    # time correction
    # time in CRS for going over the next day in UTC
    # so add 24 hours to the time to get the correct time"""
    mask = np.where(hour < hour[0])
    hour[mask] = hour[mask] + 24
    return hour
  
  def _get_date_from_url(self, url: str) -> np.datetime64:
    # get date from url
    # date is in the format of YYYYMMDD
    # eg. 20190801
    date = url.split("olympex_CRS_")[1].split("_")[0]
    np_date = np.datetime64('{}-{}-{}'.format(date[:4], date[4:6], date[6:]))
    return np_date

  def _down_vector(self, roll, pitch, head):
    x = np.sin(roll) * np.cos(head) + np.cos(roll) * np.sin(pitch) * np.sin(head)
    y = -np.sin(roll) * np.sin(head) + np.cos(roll) * np.sin(pitch) * np.cos(head)
    z = -np.cos(roll) * np.cos(pitch)
    return (x, y, z)

  def _create_zarr_dir(self):
    """Create a directory to hold zarr file
    """
    date = self._get_date_from_url(self.url)
    tempdir = 'temp/' + str(date) + '/zarr'
    if os.path.exists(tempdir):
        shutil.rmtree(tempdir)
    os.makedirs(tempdir)
    return tempdir


import json
import numpy as np
from datetime import datetime
from threading import Thread, Lock
from copy import deepcopy

from .tiles_model import tileset_json

to_rad = np.pi / 180.0
to_deg = 180.0 / np.pi

steps = [32, 16, 8, 4, 2, 1]

class PointCloud:
    def __init__(self, key, lon, lat, alt, value, time, epoch):
        self.key = key
        self.lon = lon
        self.lat = lat
        self.alt = alt
        self.time = time
        self.value = value
        self.epoch = epoch
        self.tasks = []
        self.threads = []
        for i in range(10):
            self.threads.append(Thread(target=self.worker_function))
        self.tileset_lock = Lock()
        self.tileset_json = deepcopy(tileset_json)
        self.tileset_json["root"]["boundingVolume"]["region"] = [
                        float(np.min(lon)) * to_rad,
                        float(np.min(lat)) * to_rad,
                        float(np.max(lon)) * to_rad,
                        float(np.max(lat)) * to_rad,
                        float(np.min(alt)) * to_rad,
                        float(np.max(alt)) * to_rad
                      ]
        self.tileset_json["properties"]["epoch"] = "{}Z".format(datetime.utcfromtimestamp(epoch).isoformat())


    def worker_function(self):
        while len(self.tasks) > 0:
                tile, start, end = self.tasks.pop()
                # print(tile, start, end)
                self.generate(tile, start, end)


    def start(self):
        for t in self.threads:
            t.start()


    def join(self):
        for t in self.threads:
            t.join()
        with open('{}/tileset.json'.format(self.key), mode='w+') as outfile:
            json.dump(self.tileset_json, outfile)


    def schedule_task(self, tile, start, end):
        self.tasks.append((tile, start, end))


    def generate(self, tile, start, end):
        # print(tile, start, end)
        parent_tile = self.tileset_json["root"]
        cartesian, offset, scale, cartographic, region = self.cartographic_to_cartesian(start, end)

        value = self.value[start:end]
        time = self.time[start:end]

        epoch = int(np.min(time) + self.epoch - 300)
        epoch = "{}Z".format(datetime.utcfromtimestamp(epoch).isoformat())
        end = int(np.max(time) + self.epoch + 300)
        end = "{}Z".format(datetime.utcfromtimestamp(end).isoformat())

        header_length = 28
        magic = np.string_("pnts")
        version = 1

        for step in steps:
            self.tileset_lock.acquire()
            try:
                filename = "{}_{}.pnts".format(tile, step)
                child_tile = {
                    "availability": "{}/{}".format(epoch, end),
                    "geometricError": step * 500,
                    "boundingVolume": {
                        "region": region
                    },
                    "content": {
                        "uri": filename
                    },
                    "refine": "REPLACE"
                }
                if step == 1:
                    self.tileset_json["properties"]["refined"].append(filename)
                else:
                    child_tile["children"] = []
                parent_tile["children"].append(child_tile)
                parent_tile = child_tile
            finally:
                self.tileset_lock.release()

            tile_length = 0
            feature_table_binary_byte_length = 0
            batch_table_binary_byte_length = 0
            length = value[::step].size

            feature_table_json = {
                "POINTS_LENGTH": length,
                "BATCH_LENGTH": length,
                "BATCH_ID": {
                    "byteOffset": 0,
                    "componentType": "UNSIGNED_INT"
                },
                "POSITION_QUANTIZED": {
                    "byteOffset": length * 4
                },
                "QUANTIZED_VOLUME_OFFSET": offset,
                "QUANTIZED_VOLUME_SCALE": scale
            }

            batch_table_json = {
                "value": {
                    "byteOffset": 0,
                    "componentType": "FLOAT",
                    "type": "SCALAR"
                },
                "time": {
                    "byteOffset": length * 4,
                    "componentType": "FLOAT",
                    "type": "SCALAR"
                },
                "location": {
                    "byteOffset": length * 8,
                    "componentType": "SHORT",
                    "type": "VEC3"
                }
            }

            tile_length += header_length

            feature_table_json_min = json.dumps(feature_table_json, separators=(",", ":")) + "       "
            feature_table_trim = (tile_length + len(feature_table_json_min)) % 8
            if feature_table_trim != 0:
                feature_table_json_min = feature_table_json_min[:-feature_table_trim]

            tile_length += len(feature_table_json_min)

            feature_table_binary_byte_length = length * 4 + length * 3 * 2
            tile_length += feature_table_binary_byte_length
            feature_table_padding = tile_length % 8
            if feature_table_padding != 0:
                feature_table_padding = 8 - feature_table_padding
            tile_length += feature_table_padding

            batch_table_json_min = json.dumps(batch_table_json, separators=(",", ":")) + "       "
            batch_table_trim = (tile_length + len(batch_table_json_min)) % 8
            if batch_table_trim != 0:
                batch_table_json_min = batch_table_json_min[:-batch_table_trim]

            tile_length += len(batch_table_json_min)

            batch_table_binary_byte_length = length * 4 * 2 + length * 2 * 3
            tile_length += batch_table_binary_byte_length
            batch_table_padding = tile_length % 8
            if batch_table_padding != 0:
                batch_table_padding = 8 - batch_table_padding
            tile_length += batch_table_padding

            with open('{}/{}'.format(self.key, filename), mode='wb+') as outfile:
                outfile.write(np.string_(magic).tobytes())
                outfile.write(np.uint32(version).tobytes())
                outfile.write(np.uint32(tile_length).tobytes())
                outfile.write(np.uint32(len(feature_table_json_min)).tobytes())
                outfile.write(np.uint32(feature_table_binary_byte_length + feature_table_padding).tobytes())
                outfile.write(np.uint32(len(batch_table_json_min)).tobytes())
                outfile.write(np.uint32(batch_table_binary_byte_length + batch_table_padding).tobytes())
                outfile.write(np.string_(feature_table_json_min).tobytes())
                outfile.write(np.arange(length, dtype=np.uint32).tobytes())
                outfile.write(cartesian[::step, :].tobytes())
                for _ in range(feature_table_padding):
                    outfile.write(np.string_(" ").tobytes())
                outfile.write(np.string_(batch_table_json_min).tobytes())
                outfile.write(value[::step].astype(np.float32).tobytes())
                outfile.write(time[::step].astype(np.float32).tobytes())
                outfile.write(cartographic[::step, :].tobytes())
                for _ in range(batch_table_padding):
                    outfile.write(np.string_(" ").tobytes())
                outfile.seek(0)


    def cartographic_to_cartesian(self, start, end):
        lon = self.lon[start:end]
        lat = self.lat[start:end]
        alt = self.alt[start:end]
        size = lon.size

        cartographic = np.zeros(shape=(size, 3), dtype=np.int16)
        cartographic[:, 0] = (lon * 32767 / 180).astype(np.int16)
        cartographic[:, 1] = (lat * 32767 / 180).astype(np.int16)
        cartographic[:, 2] = (alt / 10).astype(np.int16)

        lon = lon * to_rad
        lat = lat * to_rad

        radiiSquared = np.array([40680631590769, 40680631590769, 40408299984661.445], dtype=np.float64)

        N1 = np.multiply(np.cos(lat), np.cos(lon))
        N2 = np.multiply(np.cos(lat), np.sin(lon))
        N3 = np.sin(lat)

        magnitude = np.sqrt(np.square(N1) + np.square(N2) + np.square(N3))

        N1 = N1 / magnitude
        N2 = N2 / magnitude
        N3 = N3 / magnitude

        K1 = radiiSquared[0] * N1
        K2 = radiiSquared[1] * N2
        K3 = radiiSquared[2] * N3

        gamma = np.sqrt(np.multiply(N1, K1) + np.multiply(N2, K2) + np.multiply(N3, K3))

        K1 = K1 / gamma
        K2 = K2 / gamma
        K3 = K3 / gamma

        N1 = np.multiply(N1, alt)
        N2 = np.multiply(N2, alt)
        N3 = np.multiply(N3, alt)

        # x = np.multiply((N1 + K1), np.random.normal(1, .00005, N1.size))
        # y = np.multiply((N2 + K2), np.random.normal(1, .00005, N1.size))
        # z = np.multiply((N3 + K3), np.random.normal(1, .00005, N1.size))

        x = N1 + K1
        y = N2 + K2
        z = N3 + K3

        offset = [float(np.min(x)), float(np.min(y)), float(np.min(z))]

        x = x - offset[0]
        y = y - offset[1]
        z = z - offset[2]

        scale = [float(abs(np.max(x))), float(abs(np.max(y))), float(abs(np.max(z)))]

        cartesian = np.zeros(shape=(size, 3), dtype=np.uint16)
        cartesian[:, 0] = (x / scale[0] * 65535.0).astype(np.uint16)
        cartesian[:, 1] = (y / scale[1] * 65535.0).astype(np.uint16)
        cartesian[:, 2] = (z / scale[2] * 65535.0).astype(np.uint16)

        region = [
            float(np.min(lon)),
            float(np.min(lat)),
            float(np.max(lon)),
            float(np.max(lat)),
            float(np.min(alt)),
            float(np.max(alt))
        ]

        return cartesian, offset, scale, cartographic, region



import os
import zarr
import numpy as np
from .tiles_point_cloud import PointCloud

to_rad = np.pi / 180.0
to_deg = 180.0 / np.pi

def write_tiles(variable, epoch, end, zarr_location, point_cloud_folder):
    """Generates json pointcloud from a given zarr file input

    Args:
        variable (_type_): _description_
        epoch (_type_): _description_
        end (_type_): _description_
        zarr_location (string): source zarr file.
        point_cloud_folder (string): destination folder for 3d tile json file.
    """

    #out_key = f"{os.getenv('CRS_OUTPUT_FLIGHT_PATH')}/{shortname}"
    #pc_out_key = f"{output_path}/point_cloud"

    '''
    try:
        os.mkdir(out_key)
    except:
        pass
    '''

    try:
        os.mkdir(point_cloud_folder)
    except:
        pass

    # LOAD THE DATA.
    store = zarr.DirectoryStore(zarr_location)
    root = zarr.group(store=store)

    chunk_id = root["chunk_id"][:]
    num_chunks = chunk_id.shape[0]
    id = np.argmax(chunk_id[:, 1] > epoch) - 1
    start_id = chunk_id[0 if id < 0 else id, 0]
    id = num_chunks - np.argmax(chunk_id[::-1, 1] < end)
    end_id = chunk_id[id, 0] if id < num_chunks else root["time"].size - 1

    root_epoch = root.attrs["epoch"]
    location = root["location"][start_id:end_id]
    lon = location[:, 0]
    lat = location[:, 1]
    alt = location[:, 2]
    value = root["value"][variable][start_id:end_id]
    time = root["time"][start_id:end_id]

    # filter data using mask
    epoch = epoch - root_epoch # date-time
    end = end - root_epoch
    mask = np.logical_and(time >= epoch, time <= end)
    lon = lon[mask]
    lat = lat[mask]
    alt = alt[mask]
    value = value[mask]
    time = time[mask]

    # Generate Pointcloud Tileset
    point_cloud = PointCloud(point_cloud_folder, lon, lat, alt, value, time, root_epoch)

    for tile in range(int(np.ceil(time.size / 530000))):
        start_id = tile * 530000
        end_id = np.min([start_id + 530000, time.size])
        point_cloud.schedule_task(tile, start_id, end_id)

    point_cloud.start()
    point_cloud.join()